/*
   pam_bioapi - Provides a PAM-compliant interface to use in biometrically authenticating local users.

   Copyright (C) 2005 Michael R. Crusoe <michael@qrivy.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define _GNU_SOURCE
#define PAM_SM_AUTH /* we provide authentication functions */
#define PAM_SM_PASSWORD /* we provide "password" setting functions, in our case enrollment of the biometric */
#include <security/_pam_macros.h>
#include <security/pam_modules.h>
#include <bioapi.h>
#include <bioapi_schema.h>
#include <bioapi_util.h>
#include <bioapi_uuid.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/stat.h>
#include <pwd.h> /* defines 'struct passwd', and getpwnam(). */
#include <sys/types.h> /* defines 'uid_t', etc.              */

#ifdef STORE_FILE

#define DEFBIRDBPATH "/etc/bioapi/pam/"

#else

#include <sqlite3.h>

#define DEFBIRDBPATH "/etc/bioapi/pam/biopasswd.db"
#define MAXSQLSELECT 300
#define SQLCREATETABLE "CREATE TABLE biodata (id INTEGER PRIMARY KEY,uid INTEGER,uuid TEXT,header BLOB,data BLOB,signature BLOB,type INTEGER,other BLOB);"

#endif


#ifdef STORE_FILE
int readBIR(const char *BIRUUID, const char *birDbPath, const char *user, BioAPI_INPUT_BIR *BIR);
int writeBIR(BioAPI_BIR_HANDLE bir, BioAPI_HANDLE bspHandle, const char *birDbPath, const char *biruuid, const char *user);
#else
int getNextRow (sqlite3_stmt *queryResult,BioAPI_INPUT_BIR *BIR, uid_t *uid);
BioAPI_OPERATIONS_MASK getOperationsMask (const BioAPI_UUID *uuid);
int insertBIRtoDb (BioAPI_BIR_HANDLE bir, BioAPI_HANDLE bspHandle, const char *birDbPath, const char *uuidString, const char *user);
void PrintErrorCode(const char *sOperation,const BioAPI_RETURN bioReturn);
#endif
void PrintBioAPIError(const BioAPI_RETURN bRet, const char *user, const char *uuidString);

static const struct bioapi_version version = { 1, 10 };

/* Authentication functions */

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags,
                                   int argc, const char **argv) {
	int pamRet;
	BioAPI_RETURN bRet;
	const char *user;
	const char *uuidString;
	const char *birDbPath;
	const BioAPI_FAR maxFAR = 1;
	BioAPI_HANDLE bspHandle;
	BioAPI_BOOL result = BioAPI_FALSE;
	BioAPI_OPERATIONS_MASK operationsMask = -1;
  	BioAPI_INPUT_BIR birEnroll;
	/* BioAPI_BOOL precedence = BioAPI_FALSE; */
	BioAPI_FAR achievedFAR;
	BioAPI_UUID tempUuid;
	const BioAPI_UUID *uuid;
	int tries = 3;
	
	openlog("pam_bioapi", LOG_PID | LOG_CONS, LOG_AUTHPRIV);
	
	pamRet = pam_fail_delay(pamh, 100);
	
	pamRet = pam_get_user(pamh, &user, NULL);
	if (pamRet != PAM_SUCCESS) {
		if (pamRet == PAM_CONV_AGAIN) {
			pamRet = PAM_INCOMPLETE;
		} else {
			syslog(LOG_ALERT, "Error retrieving user name: %s", pam_strerror(pamh, pamRet));
			pamRet = PAM_USER_UNKNOWN;
		}
		closelog();
		return pamRet;
	}
	
	if (argc > 0) {
		uuidString = argv[0];
		if (argc > 1) {
			birDbPath = argv[1];

			// set DISPLAY variable, if defined in pam.d config
			// and not already set
			if ((argc > 2) && (getenv ("DISPLAY") == NULL)) {
				setenv ("DISPLAY", argv[2], 1);
			}
		} else {
			birDbPath = DEFBIRDBPATH;
		}
	} else {
		syslog(LOG_ALERT, "Mandatory BSP UUID not provided");
		closelog();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bRet = BioAPI_Init(&version, 0, NULL, 0, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to initilize BioAPI framework, BioAPI error 0x%x.", bRet);
		closelog();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bRet = BioAPI_GetStructuredUUID(uuidString, &tempUuid);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to parse UUID (BioAPI error 0x%x) : %s", bRet, uuidString);
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	uuid = (const BioAPI_UUID *) malloc(sizeof(BioAPI_UUID));
	if (uuid == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory.");
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}

//#ifndef STORE_FILE
	// TODO: make this function working!!! 
	// (right now it returning strange data)
	//operationsMask = getOperationsMask (uuid);
//#endif //ifndef STORE_FILE
	
	memcpy(uuid, tempUuid, sizeof(BioAPI_UUID));
	bRet = BioAPI_ModuleLoad(uuid, 0, NULL, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to load BioAPI BSP with UUID of %s, BioAPI error 0x%x.", uuidString, bRet);
		closelog();
		free(uuid);
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bRet = BioAPI_ModuleAttach(uuid, &version, &BioAPIMemoryFuncs, 0, 0, 0, 0, NULL, 0, NULL, &bspHandle);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to attach default device to BioAPI BSP with UUID of %s, BioAPI error 0x%x.", uuidString, bRet);
		closelog();
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
		return PAM_AUTHINFO_UNAVAIL;
	}
#ifdef STORE_FILE
	pamRet = readBIR(uuidString, birDbPath, user, &birEnroll);
	if (pamRet != 0) {
		closelog();
		(void)BioAPI_ModuleDetach(bspHandle);
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
		return pamRet;
	}
#else
	sqlite3 *db;
	char *zErrMsg = 0;
	int rc = -1;
	char sqlSelect[MAXSQLSELECT];
	int maskNeeded = 0;
	int authenticateOnly = 0; // false;
	uid_t uid = -1;
	sqlite3_stmt * queryResult = NULL;
	struct passwd *userPw = NULL;
  	BioAPI_BIR_HANDLE capturedTemplate, processedBir;
  	BioAPI_INPUT_BIR birCapture, inputBirProcessed;
	BioAPI_BOOL bPrecedence = BioAPI_TRUE;
	
	if (strlen(user) != 0) {
	  	// authentisation
		authenticateOnly = 1; // true;
	  	userPw = getpwnam (user);
		if (userPw == NULL) {
			syslog(LOG_ALERT, "Error while geting UID of user %s\n",user);
			closelog();
			(void)BioAPI_ModuleDetach(bspHandle);
			(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
			free(uuid);
			(void)BioAPI_Terminate();
			return PAM_USER_UNKNOWN;
		}
		uid = userPw->pw_uid;
		
		snprintf (sqlSelect, MAXSQLSELECT,"SELECT uid,header,data,signature FROM biodata WHERE uuid = '%s' AND uid = %d ORDER BY uid ASC;", uuidString, uid);
	} else {
		// identification
		snprintf (sqlSelect, MAXSQLSELECT, "SELECT uid,header,data,signature FROM biodata WHERE uuid = '%s' ORDER BY uid ASC;", uuidString);
	}
	syslog (LOG_ALERT, "DEBUG:\"%s\"\n", sqlSelect);

 	rc = sqlite3_open (birDbPath, &db);
      	if( rc ){
        	syslog(LOG_ALERT, "Can't open database: %s\n", sqlite3_errmsg(db));
        	sqlite3_close(db);
		closelog();
		(void)BioAPI_ModuleDetach(bspHandle);
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
        	return PAM_AUTHINFO_UNAVAIL;
      	}

	rc = sqlite3_prepare(db, sqlSelect, -1, &queryResult, NULL);
	if (rc == SQLITE_OK && queryResult != NULL) {
		rc = getNextRow (queryResult,&birEnroll, &uid);
		if (rc != SQLITE_ROW) {
			switch (rc) {
				case SQLITE_DONE:
					syslog(LOG_ALERT, "No biometrics data returned from database.\n");
					break;
				case -1:
					syslog(LOG_ALERT, "Can't get user ID from database.\n");
					break;
				case -2:
					syslog(LOG_ALERT, "Can't get header of BIR from database.\n");
					break;
				case -3:
					syslog(LOG_ALERT, "Can't get BiometricData of BIR from database.\n");
					break;
				case -4:
					syslog(LOG_ALERT, "Can't get signature of BIR from database.\n");
					break;
				default:
					syslog(LOG_ALERT, "Can't get Biometrics Records from database: %s\n", sqlite3_errmsg(db));
					break;
			}
			rc = sqlite3_finalize(queryResult);
			sqlite3_close(db);
			closelog();
			(void)BioAPI_ModuleDetach(bspHandle);
			(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
			free(uuid);
			(void)BioAPI_Terminate();
			return PAM_AUTHINFO_UNAVAIL;
		}

		// which operations are needed to one-to-few identification
		maskNeeded = BioAPI_CAPTURE | BioAPI_PROCESS | BioAPI_VERIFYMATCH;
		//syslog (LOG_ALERT, "DEBUG: maskNeeded = %d\n",maskNeeded);

		// See if the BSP supports all needed operations by checking
		// the operations mask
		// TODO: make it working!!!
		//if (! (operationsMask & maskNeeded)) {
		if (0) {
			if (authenticateOnly) {
#endif //ifdef STORE_FILE
				do {
					bRet = BioAPI_Verify(bspHandle, &maxFAR, NULL, NULL, &birEnroll, NULL, &result, &achievedFAR, NULL, NULL, -1, NULL);
					if ((bRet == BioAPI_OK) || (bRet == BioAPIERR_BSP_UNABLE_TO_CAPTURE)) {
						if (result == BioAPI_TRUE) {
							tries = 0;
						} else {
							tries--;
						}
					} else { 
						PrintBioAPIError(bRet,user,uuidString);
						
						closelog();
						(void)BioAPI_ModuleDetach(bspHandle);
						(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
						free(uuid);
						(void)BioAPI_Terminate();
						return PAM_AUTHINFO_UNAVAIL;
					}

				} while (tries > 0);
	
#ifndef STORE_FILE // continue of sqlite3 store implementation
			} else {
				syslog (LOG_ALERT, "Operations for identification doesn't supported by module %s\n", uuid);
				rc = sqlite3_finalize(queryResult);
				sqlite3_close(db);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			}
		} else {
			// one-to-few identification
			if ((bRet = BioAPI_Capture (bspHandle, BioAPI_PURPOSE_VERIFY,
						&capturedTemplate, -1, NULL)) != BioAPI_OK) {
				syslog (LOG_ALERT,"Error while calling BioAPI_Capture");
				free(birEnroll.InputBIR.BIR);
				rc = sqlite3_finalize(queryResult);
				sqlite3_close(db);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			}
			
			/*
			 * Do I need this? 
			BioAPI_BIR_HEADER birHeader;
			if ((bRet = BioAPI_GetHeaderFromHandle(bspHandle, capturedTemplate, &birHeader)) != BioAPI_OK) {
				syslog (LOG_ALERT,"BioAPI_GetHeaderFromHandle");
				//free(birEnroll.InputBIR.BIR);
				rc = sqlite3_finalize(queryResult);
				sqlite3_close(db);
				return PAM_AUTHINFO_UNAVAIL;
			}
			*/

			birCapture.Form = BioAPI_BIR_HANDLE_INPUT;
			birCapture.InputBIR.BIRinBSP = &capturedTemplate;
			if ((bRet = BioAPI_Process(bspHandle, &birCapture, &processedBir)) != BioAPI_OK) {
				syslog (LOG_ALERT,"BioAPI_Process",bRet);
				free(birEnroll.InputBIR.BIR);
				rc = sqlite3_finalize(queryResult);
				sqlite3_close(db);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHINFO_UNAVAIL;
			}
			inputBirProcessed.Form = BioAPI_BIR_HANDLE_INPUT;
			inputBirProcessed.InputBIR.BIRinBSP = &processedBir;

			do {

				syslog(LOG_INFO, "DEBUG testing: uid = %d, uuid = %s", uid, uuidString);
				
				bRet = BioAPI_VerifyMatch(bspHandle, &maxFAR, NULL,
						     &bPrecedence, &inputBirProcessed, &birEnroll, NULL,
						     &result, &achievedFAR, NULL, NULL);
				if (bRet != BioAPI_OK) {
						PrintBioAPIError(bRet,user,uuidString);
						
						free(birEnroll.InputBIR.BIR);
						rc = sqlite3_finalize(queryResult);
						sqlite3_close(db);
						closelog();
						(void)BioAPI_ModuleDetach(bspHandle);
						(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
						free(uuid);
						(void)BioAPI_Terminate();
						return PAM_AUTHINFO_UNAVAIL;
				}

				syslog(LOG_INFO, "DEBUG VerifyMatch result = %d", result);
				
				if (result == BioAPI_TRUE) {
					// we have found correct identity
					if (! authenticateOnly) {
						// set user
						userPw = getpwuid (uid);
						if (userPw == NULL) {
							syslog(LOG_ALERT, "Biometrics records match but can't get username for uid = %d\n",uid);
							rc = sqlite3_finalize(queryResult);
							sqlite3_close(db);
							closelog();
							(void)BioAPI_ModuleDetach(bspHandle);
							(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
							free(uuid);
							(void)BioAPI_Terminate();
							return PAM_AUTHINFO_UNAVAIL;
						}
						user = userPw->pw_name;
						pamRet = pam_set_item (pamh, PAM_USER, user);
						if (pamRet != PAM_SUCCESS) {
							syslog(LOG_ALERT, "Can't set username %s as PAM item!\n",user);
						}
					}
					break;
				}

				rc = getNextRow (queryResult,&birEnroll, &uid);
				if (rc == SQLITE_MISUSE) {
					syslog(LOG_ALERT, "Unexpected error since identificating user!\n");
					rc = sqlite3_finalize(queryResult);
					sqlite3_close(db);
					closelog();
					(void)BioAPI_ModuleDetach(bspHandle);
					(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
					free(uuid);
					(void)BioAPI_Terminate();
					return PAM_AUTHINFO_UNAVAIL;
				}
				if (rc == SQLITE_ERROR) {
					syslog(LOG_ALERT, "SQLite error: %s\n",sqlite3_errmsg(db));
					rc = sqlite3_finalize(queryResult);
					sqlite3_close(db);
					closelog();
					(void)BioAPI_ModuleDetach(bspHandle);
					(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
					free(uuid);
					(void)BioAPI_Terminate();
					return PAM_AUTHINFO_UNAVAIL;
				}
			} while (rc != SQLITE_DONE); // all rows checked

		}
	}
      	
	rc = sqlite3_finalize(queryResult);
	sqlite3_close(db);
#endif //ifdef STORE_FILE

	if (result == BioAPI_TRUE) {
		syslog(LOG_INFO, "User %s successfully biometrically authenticated using BSP with UUID of %s", user, uuid);
		pamRet = PAM_SUCCESS;
	} else {
		syslog(LOG_ALERT, "User %s failed to biometrically authenticate using BSP with UUID of %s", user, uuid);
		pamRet = PAM_AUTH_ERR;
	}
	(void)BioAPI_ModuleDetach(bspHandle);
	(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
	free(uuid);
	(void)BioAPI_Terminate();
	closelog();
	return pamRet;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags,
			      int argc, const char **argv) {
	return PAM_SUCCESS;
}

/* "Password" functions (in our case, enrollment of the biometric) */

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv) {
	int pamRet;
	BioAPI_RETURN bRet;
	const char *user;	
	const char *uuidString;
	const char *birDbPath;
	BioAPI_UUID tempUuid;
	const BioAPI_UUID *uuid;
	BioAPI_HANDLE bspHandle;
	BioAPI_BIR_HANDLE bir;
	int tries;

	openlog("pam_bioapi", LOG_PID | LOG_CONS, LOG_AUTHPRIV);
	
	pamRet = pam_fail_delay(pamh, 100);
	if ((flags & PAM_PRELIM_CHECK) == PAM_PRELIM_CHECK) {
		closelog();
		return PAM_SUCCESS; /* No remote resource to check */
	}
	
	pamRet = pam_get_user(pamh, &user, NULL);
	if (pamRet != PAM_SUCCESS || strlen(user) == 0) {
		if (pamRet == PAM_CONV_AGAIN) {
			pamRet = PAM_INCOMPLETE;
		} else {
			syslog(LOG_ALERT, "Error retrieving user name: %s", pam_strerror(pamh, pamRet));
			pamRet = PAM_USER_UNKNOWN;
		}
		closelog();
		return pamRet;
	}
	
	if (argc > 0) {
		uuidString = argv[0];
		if (argc > 1) {
			birDbPath = argv[1];
		} else {
			birDbPath = DEFBIRDBPATH;
		}
	} else { 
		syslog(LOG_ALERT, "Mandatory BSP UUID not provided");
		closelog();
		return PAM_AUTHTOK_ERR;
	}
	
	bRet = BioAPI_Init(&version, 0, NULL, 0, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to initilize BioAPI framework, BioAPI error #:%x.", bRet);
		closelog();
		return PAM_AUTHTOK_ERR;
	}
	
	bRet = BioAPI_GetStructuredUUID(uuidString, &tempUuid);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to parse UUID (BioAPI error #:%x) : %s", bRet, uuidString);
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}

	uuid = (const BioAPI_UUID *) malloc(sizeof(BioAPI_UUID));
	if (uuid == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory.");
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}
	
	memcpy(uuid, tempUuid, sizeof(BioAPI_UUID));
	bRet = BioAPI_ModuleLoad(uuid, 0, NULL, NULL);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to load BioAPI BSP with UUID of %s, BioAPI error #%x.", uuidString, bRet);
		closelog();
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}

	bRet = BioAPI_ModuleAttach(uuid, &version, &BioAPIMemoryFuncs, 0, 0, 0, 0, NULL, 0, NULL, &bspHandle);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to attach default device to BioAPI BSP with UUID of %s, BioAPI error #%x.", uuidString, bRet);
		closelog();
		(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
		free(uuid);
		(void)BioAPI_Terminate();
		return PAM_AUTHTOK_ERR;
	}
	
	tries = 3;
	do {
		bRet = BioAPI_Enroll (bspHandle, BioAPI_PURPOSE_ENROLL_FOR_VERIFICATION_ONLY, NULL, &bir, NULL, -1, NULL);
		switch (bRet) {
			case BioAPI_OK:
				tries = 0;
#ifdef STORE_FILE
				pamRet = writeBIR (bir, bspHandle, birDbPath, uuidString, user);
#else
				pamRet = insertBIRtoDb (bir, bspHandle, birDbPath, uuidString, user);
#endif
				break;
			case BioAPIERR_BSP_UNABLE_TO_CAPTURE:
			case BioAPIERR_BSP_TIMEOUT_EXPIRED:
				tries--;
				break;
			case BioAPIERR_BSP_TOO_MANY_HANDLES:
			case BioAPIERR_BSP_PURPOSE_NOT_SUPPORTED:
			case BioAPIERR_BSP_RECORD_NOT_FOUND:
			default:
				syslog(LOG_ALERT, "Unable to enroll user %s using BSP with UUID of %s, BioAPI error #%x.", user, uuidString, bRet);
				closelog();
				(void)BioAPI_ModuleDetach(bspHandle);
				(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
				free(uuid);
				(void)BioAPI_Terminate();
				return PAM_AUTHTOK_ERR;
		}
	} while (tries > 0);
	(void)BioAPI_ModuleDetach(bspHandle);
	(void)BioAPI_ModuleUnload(uuid, NULL, NULL);
	free(uuid);
	(void)BioAPI_Terminate();
	closelog();
	if (bRet != PAM_SUCCESS) {
		return PAM_AUTHTOK_ERR;
	} else {
		return bRet;
	}
}
#ifdef STORE_FILE
int readBIR(const char *biruuid, const char *birDbPath, const char *user, BioAPI_INPUT_BIR *bir) {
	FILE *inputFile;
	char *filename = (char *) malloc (strlen(birDbPath) + strlen(biruuid) + 1 + strlen(user) + 4 + 1);
	size_t bytes, bytesRead;
	BioAPI_BIR *tempBir;

	strcpy(filename, birDbPath);
	strcat(filename, biruuid);
	strcat(filename, "/");
	strcat(filename, user);
	strcat(filename, ".bir");
	
	inputFile = fopen (filename, "r");
	if (inputFile == NULL) {
		syslog(LOG_ALERT, "Unable to open biometric identification record, \"%s\", for user \"%s\"", filename, user);
		free(filename);
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	free(filename);
	
	bytes = sizeof(BioAPI_BIR_HEADER);
	tempBir = (BioAPI_BIR *)malloc(sizeof(BioAPI_BIR));
	if (tempBir == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory for BIR representation (tempBir).");
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bytesRead = fread(&(tempBir->Header), bytes, 1, inputFile);
//	bytes = LocalEndian4 (tempBir->Header.Length) - sizeof(BioAPI_DATA); 
	bytes = LocalEndian4 (tempBir->Header.Length) - sizeof(BioAPI_BIR_HEADER); 
	tempBir->BiometricData = (BioAPI_BIR_BIOMETRIC_DATA_PTR)malloc(bytes);
	if (tempBir->BiometricData == 0) {
		syslog(LOG_ALERT, "Unable to allocate memory for BiometricData in BIR representation (tempBir->BiometricData).");
		free(tempBir);
		return PAM_AUTHINFO_UNAVAIL;
	}
	
	bytesRead = fread(tempBir->BiometricData, bytes, 1, inputFile);
	bytes = 0;
	bytesRead = fread(&bytes, 4, 1, inputFile);
	
	// if (&bytesRead != 0 && &bytes != 0) { ?????
	if (bytesRead != 0 && bytes != 0) {
		tempBir->Signature = (BioAPI_DATA_PTR)malloc(sizeof(BioAPI_DATA));
		if (tempBir->Signature == 0) {
			syslog(LOG_ALERT, "Unable to allocate memory for BIR signature.");
			free(tempBir->BiometricData);
			free(tempBir);
			return PAM_AUTHINFO_UNAVAIL;
		}	
		tempBir->Signature->Length = (uint32)bytes;
		tempBir->Signature->Data = (uint8 *)malloc(bytes);
		if (tempBir->Signature->Data == 0) {
			syslog(LOG_ALERT, "Unable to allocate memory for BIR signature data.");
			free(tempBir->BiometricData);
			free(tempBir->Signature);
			free(tempBir);
			return PAM_AUTHINFO_UNAVAIL;
		}	
		bytesRead = fread(tempBir->Signature->Data, bytes, 1, inputFile);
	}
	(void)fclose(inputFile);
	bir->InputBIR.BIR = tempBir;
	bir->Form = BioAPI_FULLBIR_INPUT;
	return PAM_SUCCESS;
}

int writeBIR(BioAPI_BIR_HANDLE bir, BioAPI_HANDLE bspHandle, const char *birDbPath, const char *biruuid, const char *user) {
	FILE *outputFile;
	char *filename = (char*) malloc(strlen(birDbPath) + strlen(biruuid) + 1 + strlen(user) + 4 + 1);
	struct stat status;
	BioAPI_BIR_PTR birData;
	BioAPI_RETURN bRet;

	strcpy(filename, birDbPath);
	(void)stat(filename, &status); /* Does the birDb directory exist? */
	if (errno == ENOENT) { /* No. */
		if (mkdir(filename, S_IRWXU | S_IRWXG) == -1) { /* Try to make it */
			syslog(LOG_ALERT, "Unable to create BIR database directory, \"%s\"", filename);
			free(filename);
			return PAM_AUTHTOK_ERR; /* But fail out if we can't. */
		}
	}
	
	strcat(filename, biruuid);
	(void)stat(filename, &status); /* Does the BSP specific directory exist? */
	if (errno == ENOENT) { /* No. */
		if (mkdir(filename, S_IRWXU | S_IRWXG) == -1) { /* Try to make it */
			syslog(LOG_ALERT, "Unable to create BSP-specific subdirectory in BIR database directory, \"%s\"", filename);
			free(filename);
			return PAM_AUTHTOK_ERR; /* But fail out if we can't */
		}
	}
	
	strcat(filename, "/");
	strcat(filename, user);
	strcat(filename, ".bir");
	
	birData = NULL;
	bRet = BioAPI_GetBIRFromHandle(bspHandle, bir, &birData);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to write biometric identification record, \"%s\": BioAPI error #%x", filename, bRet);
		free(filename);
		return PAM_AUTHTOK_ERR;
	}
	
	outputFile = fopen(filename, "w+");
	free(filename);

	if (outputFile == NULL) {
		syslog(LOG_ALERT, "Unable to open BIR for writing, \"%s\"", filename);
		return PAM_AUTHTOK_ERR;
	}
	
	(void)fwrite(&(birData->Header), sizeof(BioAPI_BIR_HEADER), 1, outputFile);
	(void)fwrite(birData->BiometricData, birData->Header.Length - sizeof(BioAPI_BIR_HEADER), 1, outputFile);

	if (birData->Signature) {
		(void)fwrite(&(birData->Signature->Length), 4, 1, outputFile);
		(void)fwrite(birData->Signature->Data, (size_t)birData->Signature->Length, 1, outputFile);
	}

	(void)fclose(outputFile);
	free(birData->BiometricData);

	if (birData->Signature) {
		free(birData->Signature->Data);
		free(birData->Signature);
	}
	
	free(birData);

	return PAM_SUCCESS;
}


#else

/* in: queryResult, out: BIR, out: uid */
int getNextRow (sqlite3_stmt *queryResult,BioAPI_INPUT_BIR *BIR, uid_t *uid){
	void * data = NULL;
	int rc = SQLITE_BUSY;
	int sqltype = SQLITE_NULL;
	BioAPI_BIR *tempBir;

	while (rc == SQLITE_BUSY) // until database is locked
	    	rc = sqlite3_step (queryResult);
		
	if (rc  == SQLITE_ROW) {
		// get uid from query
		if (sqlite3_column_type (queryResult, 0) == SQLITE_INTEGER) {
			*uid = sqlite3_column_int (queryResult, 0);
		} else return -1;
		
		tempBir = (BioAPI_BIR *)malloc(sizeof(BioAPI_BIR));
		// get BIR from query
		if (sqlite3_column_type (queryResult, 1) == SQLITE_BLOB) {
			data = (void *) sqlite3_column_blob (queryResult, 1);
		} else { free (tempBir); return -2; }
		memcpy (&(tempBir->Header), (BioAPI_BIR_HEADER *)data, sizeof (BioAPI_BIR_HEADER));
		//free (data);

		if (sqlite3_column_type (queryResult, 2) == SQLITE_BLOB) {
			data = (void *) sqlite3_column_blob (queryResult, 2);
		} else { free (tempBir); return -3; }
		tempBir->BiometricData = (BioAPI_BIR_BIOMETRIC_DATA_PTR) data;

		sqltype = sqlite3_column_type (queryResult, 3);
		if (sqltype == SQLITE_BLOB) {
			data = (void *) sqlite3_column_blob (queryResult, 3);
			
			tempBir->Signature = (BioAPI_DATA_PTR)malloc(sizeof(BioAPI_DATA));
			if (tempBir->Signature == NULL) {
				syslog(LOG_ALERT, "Unable to allocate memory for BIR signature.");
				free(tempBir->BiometricData);
				free(tempBir);
				return PAM_AUTHINFO_UNAVAIL;
			}
			tempBir->Signature->Length = (uint32) sqlite3_column_bytes (queryResult, 3);
			tempBir->Signature->Data = data;
		} else if (sqltype == SQLITE_NULL) {
			tempBir->Signature = NULL;
		} else { free (tempBir); return -4; }

		BIR->InputBIR.BIR = tempBir;
		BIR->Form = BioAPI_FULLBIR_INPUT;

	}
	return rc;
}

BioAPI_OPERATIONS_MASK getOperationsMask (const BioAPI_UUID *uuid) {
	BioAPI_OPERATIONS_MASK ret = 0;
	BioAPI_BSP_SCHEMA *BspSchemaArray = NULL;
	BioAPI_BSP_SCHEMA *CurrSchema = NULL;
	uint32 ArraySize, ElementsNeeded, NumElementsReturned, i;
	BioAPI_RETURN bioReturn = 0;


        bioReturn = BioAPI_EnumModules(NULL, 0, &ElementsNeeded, &NumElementsReturned);
        if(bioReturn != BioAPI_OK) 
        { 
                if (bioReturn == BioAPIERR_H_FRAMEWORK_FUNCTION_FAILED) 
                {
                        syslog (LOG_ALERT, "No BSPs have been installed\n");
                }
                else
                {
                        PrintErrorCode("BioAPI_EnumModules",bioReturn);
                }
                return -1;      
        }
        
        ArraySize = ElementsNeeded;
        BspSchemaArray = (BioAPI_BSP_SCHEMA *) malloc(ElementsNeeded * sizeof(BioAPI_BSP_SCHEMA));
        if(BspSchemaArray == NULL) 
        { 
                syslog (LOG_ALERT, "Unable to allocate BSP list\n");
                return -1;
        }
        bioReturn = BioAPI_EnumModules(BspSchemaArray, ArraySize, &ElementsNeeded, &NumElementsReturned);
        if(bioReturn != BioAPI_OK)
        { 
                free(BspSchemaArray);
                PrintErrorCode("BioAPI_EnumModules",bioReturn);
                return -1;
        }
        
        CurrSchema = BspSchemaArray;
        for(i = 0; i < NumElementsReturned; i++) 
        { 
                if(!memcmp(CurrSchema->ModuleId,uuid,sizeof(uuid)))
                {
		    break;
                }   
                CurrSchema++;
        } 
	
	ret = CurrSchema->Operations;

	free(BspSchemaArray);
        
	//syslog (LOG_ALERT, "DEBUG: CurrSchema->Operations = %d\n",ret);

	return ret;
}

int insertBIRtoDb (BioAPI_BIR_HANDLE bir, BioAPI_HANDLE bspHandle, const char *birDbPath, const char *uuidString, const char *user) {
	struct stat status;
	BioAPI_BIR_PTR birData;
	BioAPI_RETURN bRet;
	sqlite3 *db;
	char *zErrMsg = 0;
	int rc;
 	int isNewFile = 0; //false
 	int type = 0;
	uid_t uid = -1;
	char sqlCommand [MAXSQLSELECT];
	sqlite3_stmt * queryResult = NULL;
	struct passwd *userPw = NULL;

	(void)stat(birDbPath, &status); 
	if (! S_ISREG (status.st_mode)) { // Is path provided regular file?
		// we have to make new
		isNewFile = 1;
	}
	
	birData = NULL;
	bRet = BioAPI_GetBIRFromHandle(bspHandle, bir, &birData);
	if (bRet != BioAPI_OK) {
		syslog(LOG_ALERT, "Unable to write biometric identification record, \"%s\": BioAPI error #%x", birDbPath, bRet);
		return PAM_AUTHTOK_ERR;
	}
	
	// open database
	rc = sqlite3_open (birDbPath, &db);
      	if( rc ){
		sqlite3_close(db);
        	syslog(LOG_ALERT, "Can't open database: %s\n", sqlite3_errmsg(db));
        	return PAM_AUTHTOK_ERR;
      	}

	if ( isNewFile) {
		rc = sqlite3_exec (db, SQLCREATETABLE, NULL, NULL, &zErrMsg);
		syslog (LOG_ALERT, "DEBUG: %s", SQLCREATETABLE);
		if( rc != SQLITE_OK ){
			syslog(LOG_ALERT, "SQL error while creating table biodata: %s\n", zErrMsg);
			sqlite3_close(db);
		        /* This will free zErrMsg if assigned */
			if (zErrMsg)
				free(zErrMsg);
			return PAM_AUTHTOK_ERR;
		}
	}

  	userPw = getpwnam (user);
	if (userPw == NULL) {
		syslog(LOG_ALERT, "Error while geting UID of user %s\n",user);
		sqlite3_close(db);
		return PAM_USER_UNKNOWN;
	}
	uid = userPw->pw_uid;
	
	//TODO: get type from app by get_pam_item()
	snprintf (sqlCommand, MAXSQLSELECT,"INSERT INTO biodata (uid,uuid,header,data,signature,type) VALUES (%d,'%s',?,?,?,%d);", uid, uuidString, type);
	//snprintf (sqlCommand, MAXSQLSELECT,"INSERT INTO biodata (uid,uuid,type) VALUES (%d,'%s',0);", uid, uuidString);
	
	syslog (LOG_ALERT, "DEBUG:\"%s\"\n", sqlCommand);
	
	rc = sqlite3_prepare (db, sqlCommand, -1, &queryResult, 0);
	if ((rc != SQLITE_OK) || (queryResult == NULL)) {
		syslog(LOG_ALERT, "SQL error while inserting data into table biodata: %s\n", sqlite3_errmsg(db));
		rc = sqlite3_finalize(queryResult);
		sqlite3_close(db);
		return PAM_AUTHTOK_ERR;
	}

	//(void)fwrite(&(birData->Header), sizeof(BioAPI_BIR_HEADER), 1, outputFile);
	rc = sqlite3_bind_blob (queryResult, 1, &(birData->Header), sizeof(BioAPI_BIR_HEADER), SQLITE_STATIC);
	if (rc != SQLITE_OK)  {
		syslog(LOG_ALERT, "SQL error while inserting Header into table biodata: %s\n", sqlite3_errmsg(db));
		rc = sqlite3_finalize(queryResult);
		sqlite3_close(db);
		return PAM_AUTHTOK_ERR;
	}

	//(void)fwrite(birData->BiometricData, birData->Header.Length - sizeof(BioAPI_BIR_HEADER), 1, outputFile);
	rc = sqlite3_bind_blob (queryResult, 2, birData->BiometricData, birData->Header.Length - sizeof(BioAPI_BIR_HEADER), free);
	if (rc != SQLITE_OK)  {
		syslog(LOG_ALERT, "SQL error while inserting BiometricData into table biodata: %s\n", sqlite3_errmsg(db));
		rc = sqlite3_finalize(queryResult);
		sqlite3_close(db);
		return PAM_AUTHTOK_ERR;
	}

	if (birData->Signature) {
		//(void)fwrite(&(birData->Signature->Length), 4, 1, outputFile);
		//(void)fwrite(birData->Signature->Data, (size_t)birData->Signature->Length, 1, outputFile);
		rc = sqlite3_bind_blob (queryResult, 3, birData->Signature->Data, (size_t)birData->Signature->Length, free);
		if (rc != SQLITE_OK)  {
			syslog(LOG_ALERT, "SQL error while inserting Signature->Data into table biodata: %s\n", sqlite3_errmsg(db));
			rc = sqlite3_finalize(queryResult);
			sqlite3_close(db);
			return PAM_AUTHTOK_ERR;
		}
		
		//free(birData->Signature);
	} else {
		rc = sqlite3_bind_blob (queryResult, 3, NULL, 0, SQLITE_STATIC);
		if (rc != SQLITE_OK)  {
			syslog(LOG_ALERT, "SQL error while inserting Signature into table biodata: %s\n", sqlite3_errmsg(db));
			rc = sqlite3_finalize(queryResult);
			sqlite3_close(db);
			return PAM_AUTHTOK_ERR;
		}
	}

	rc = sqlite3_step(queryResult);
	if (rc != SQLITE_DONE)  {
		syslog(LOG_ALERT, "SQL error while inserting BiometricData into table biodata: %s\n", sqlite3_errmsg(db));
		rc = sqlite3_finalize(queryResult);
		sqlite3_close(db);
		return PAM_AUTHTOK_ERR;
	}

	//free(birData);
	
	
	rc = sqlite3_finalize(queryResult);
	if (rc != SQLITE_OK)  {
		syslog(LOG_ALERT, "SQL error while finalizing insertion of data into table biodata: %s\n", sqlite3_errmsg(db));
		rc = sqlite3_finalize(queryResult);
		sqlite3_close(db);
		return PAM_AUTHTOK_ERR;
	}
	sqlite3_close(db);

	return PAM_SUCCESS;
}


void PrintErrorCode(const char *sOperation,const BioAPI_RETURN bioReturn) 
{
        syslog (LOG_ALERT, "%s failed, BioAPI Error Code: %ld (0x%lx)\n", sOperation,bioReturn, bioReturn);
}

#endif //ifdef STORE_FILE

void PrintBioAPIError(const BioAPI_RETURN bRet, const char *user, const char *uuidString) {
	switch(bRet) {
		case BioAPIERR_BSP_BIR_SIGNATURE_FAILURE:
			syslog(LOG_ALERT, "Error in BIR signature verification while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
			break;
		case BioAPIERR_BSP_TOO_MANY_HANDLES:
			syslog(LOG_ALERT, "Ran out of handles while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
			break;
		case BioAPIERR_BSP_INCONSISTENT_PURPOSE:
			syslog(LOG_ALERT, "BSP Error: Inconsistent Purpose -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
			break;
		case BioAPIERR_BSP_RECORD_NOT_FOUND:
			syslog(LOG_ALERT, "BSP Error: Record not found -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
			break;
                case BioAPI_ERRCODE_INTERNAL_ERROR:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INTERNAL_ERROR -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_MEMORY_ERROR:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_MEMORY_ERROR -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INVALID_POINTER:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INVALID_POINTER -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INVALID_INPUT_POINTER:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INVALID_INPUT_POINTER -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INVALID_OUTPUT_POINTER:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INVALID_OUTPUT_POINTER -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_FUNCTION_NOT_IMPLEMENTED:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_FUNCTION_NOT_IMPLEMENTED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_OS_ACCESS_DENIED:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_OS_ACCESS_DENIED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INVALID_UUID:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INVALID_UUID -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INCOMPATIBLE_VERSION:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INCOMPATIBLE_VERSION -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INVALID_DATA:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INVALID_DATA -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPI_ERRCODE_INVALID_DB_HANDLE:
                        syslog(LOG_ALERT, "returned BioAPI_ERRCODE_INVALID_DB_HANDLE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_INVALID_MODULE_HANDLE:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_INVALID_MODULE_HANDLE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_NOT_INITIALIZED:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_NOT_INITIALIZED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_MODULE_LOAD_FAILED:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_MODULE_LOAD_FAILED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_MODULE_UNLOAD_FAILED:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_MODULE_UNLOAD_FAILED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_LIB_REF_NOT_FOUND:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_LIB_REF_NOT_FOUND -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_INVALID_MODULE_FUNCTION_TABLE:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_INVALID_MODULE_FUNCTION_TABLE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_MODULE_NOT_LOADED:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_MODULE_NOT_LOADED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_H_FRAMEWORK_INVALID_DEVICE_ID:
                        syslog(LOG_ALERT, "returned BioAPIERR_H_FRAMEWORK_INVALID_DEVICE_ID -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INTERNAL_ERROR:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INTERNAL_ERROR -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_POINTER:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_POINTER -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_INPUT_POINTER:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_INPUT_POINTER -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_OUTPUT_POINTER:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_OUTPUT_POINTER -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_FUNCTION_NOT_IMPLEMENTED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_FUNCTION_NOT_IMPLEMENTED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_OS_ACCESS_DENIED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_OS_ACCESS_DENIED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_FUNCTION_FAILED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_FUNCTION_FAILED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_DATA:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_DATA -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_DB_HANDLE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_DB_HANDLE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_CAPTURE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_CAPTURE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_TIMEOUT_EXPIRED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_TIMEOUT_EXPIRED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_BIR:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_BIR -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_WRAP_PAYLOAD:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_WRAP_PAYLOAD -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_NO_INPUT_BIRS:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_NO_INPUT_BIRS -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNSUPPORTED_FORMAT:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNSUPPORTED_FORMAT -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_IMPORT:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_IMPORT -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_FUNCTION_NOT_SUPPORTED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_FUNCTION_NOT_SUPPORTED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_BIR_NOT_FULLY_PROCESSED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_BIR_NOT_FULLY_PROCESSED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_PURPOSE_NOT_SUPPORTED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_PURPOSE_NOT_SUPPORTED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_OPEN_DATABASE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_OPEN_DATABASE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_DATABASE_IS_LOCKED:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_DATABASE_IS_LOCKED -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_DATABASE_DOES_NOT_EXIST:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_DATABASE_DOES_NOT_EXIST -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_DATABASE_ALREADY_EXISTS:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_DATABASE_ALREADY_EXISTS -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_DATABASE_NAME:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_DATABASE_NAME -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_CURSOR_IS_INVALID:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_CURSOR_IS_INVALID -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_DATABASE_IS_OPEN:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_DATABASE_IS_OPEN -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_INVALID_ACCESS_REQUEST:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_INVALID_ACCESS_REQUEST -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_END_OF_DATABASE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_END_OF_DATABASE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_CREATE_DATABASE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_CREATE_DATABASE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_CLOSE_DATABASE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_CLOSE_DATABASE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
                case BioAPIERR_BSP_UNABLE_TO_DELETE_DATABASE:
                        syslog(LOG_ALERT, "returned BioAPIERR_BSP_UNABLE_TO_DELETE_DATABASE -- while authenticating user \"%s\" using BSP with UUID of %s.", user, uuidString);
                        break;
		default:
			syslog(LOG_ALERT, "Unanticipated BioAPI error 0x%x -- while authenticating user \"%s\" using BSP with UUID of %s", bRet, user, uuidString);
	}
}
